# -*- coding: utf-8 -*-

# FogLAMP_BEGIN
# See: https://foglamp.dianomic.com/
# FOGLAMP_END

""" Kafka Consumer """

import asyncio
import copy
import json
import logging
import uuid, sys, os

from foglamp.common import logger
from foglamp.plugins.common import utils
from foglamp.services.south import exceptions
from foglamp.services.south.ingest import Ingest
from foglamp.common.common import _FOGLAMP_ROOT, _FOGLAMP_DATA
import async_ingest

from kafka import KafkaConsumer
import threading

__author__ = "Mohit Singh Tomar"
__copyright__ = "Copyright (c) 2023 Dianomic Systems, Inc."
__license__ = "Apache 2.0"
__version__ = "${VERSION}"

_LOGGER = logger.setup(__name__)

c_callback = None
c_ingest_ref = None

_DEFAULT_CONFIG = {
    'plugin': {
        'description': 'Kafka Consumer South Plugin',
        'type': 'string',
        'default': 'kafka_python',
        'readonly': 'true'
    },
    'brokerHost': {
        'description': 'Hostname or IP address of the broker to connect to',
        'type': 'string',
        'default': 'localhost',
        'order': '1',
        'displayName': 'Kafka Broker host',
        'mandatory': 'true'
    },
    'brokerPort': {
        'description': 'The network port of the broker to connect to',
        'type': 'integer',
        'default': '9092',
        'order': '2',
        'displayName': 'Kafka Broker Port',
        'mandatory': 'true'
    },
    'group_id': {
        'description': 'The name of the consumer group to join for dynamic partition assignment and to use for fetching and committing offsets.',
        'type': 'string',
        'default': 'FogLAMP',
        'order': '3',
        'displayName': 'Kafka Group ID',
        'mandatory': 'true'
    },
    'auto_offset_reset': {
        'description': 'A policy for resetting offsets on OffsetOutOfRange errors. earliest: will move to the oldest available message,'
                        'latest: will move to the most recent.',
        'type': 'enumeration',
        'default': 'earliest',
        'order': '4',
        'displayName': 'Offset Reset',
        'options': ['earliest', 'latest'],
    },
    'enable_auto_commit': {
        'description': 'If True, the consumer’s offset will be periodically committed in the background.',
        'type': 'boolean',
        'default': 'true',
        'order': '5',
        'displayName': 'Auto Commit'
    },
    'asset': {
        'description': "Asset that will be created",
        'type': 'string',
        'default': 'FogLAMP',
        'order': '6',
        'displayName': 'Default Asset Name',
        'mandatory': 'true',
        'group': 'Asset Mapping' 
    },
    'topic_assets_map': {
        'description': 'JSON array of objects, where each object contains a topic to be subscribed and optional asset & datapoint name to map to readings',
        'type': 'JSON',
        'default': '{"map": [{"topic": "t", "asset": "t_a", "datapoint": "t_d"}]}',
        'order': '7',
        'displayName': 'Asset Map',
        'mandatory': 'true',
        'group': 'Asset Mapping' 
    },
    'heartbeat_interval_ms': {
        'description': 'Maximum period in milliseconds allowed between communications with the broker. If no other messages are being exchanged, '
                        'this controls the rate at which the client will send ping messages to the broker.',
        'type': 'integer',
        'default': '3000',
        'order': '8',
        'displayName': 'Keep Alive Interval',
        'group': 'Kafka Connection'
    },
    'request_timeout_ms': {
        'description': 'Client request timeout in milliseconds.',
        'type': 'integer',
        'default': '305000',
        'order': '9',
        'displayName': 'Request Timeout',
        'group': 'Kafka Connection'
    },
    'reconnect_backoff_ms': {
        'description': 'The amount of time in milliseconds to wait before attempting to reconnect to a given host.',
        'type': 'integer',
        'default': '50',
        'order': '10',
        'displayName': 'Backoff Time',
        'group': 'Kafka Connection'
    },
    'reconnect_backoff_max_ms': {
        'description': 'The maximum amount of time in milliseconds to backoff/wait when reconnecting to a broker that has repeatedly failed to connect.',
        'type': 'integer',
        'default': '1000',
        'order': '11',
        'displayName': 'Max Backoff Time',
        'group': 'Kafka Connection'
    },
    'timeout_ms': {
        'description': 'The maximum time in milliseconds to wait for records or messages to process before returning.',
        'type': 'integer',
        'default': '1000',
        'minimum': '0',
        'order': '12',
        'displayName': 'Message Polling Timeout',
        'group': 'Kafka Connection'
    },
    'authentication': {
        'description': 'Security Protocol or Authentication method for connecting with Kafka Broker',
        'type': 'enumeration',
        'default': 'PLAINTEXT',
        'order': '13',
        'displayName': 'Security Protocol',
        'options': ['PLAINTEXT', 'SASL_PLAINTEXT', 'SSL', 'SASL_SSL'],
        'group': 'Kafka Authentication' 
    },
    'sasl_mechanism': {
      'description': 'Security mechanism for connecting with Kafka Broker',
      'type': 'enumeration',
      'default': 'PLAIN',
      'order': '14',
      'displayName': 'SASL Mechanism',
      'options': ['PLAIN', 'GSSAPI', 'OAUTHBEARER', "SCRAM-SHA-256", "SCRAM-SHA-512"],  # TODO Add functionlity for connecting with GSSAPI, OAUTHBEARER, SCRAM-SHA-256, SCRAM-SHA-512 also.
      'validity': "authentication == \'SASL_PLAINTEXT\' || authentication == \"SASL_SSL\"",
      'group': 'Kafka Authentication'
    },
    'username': {
        'description': 'Username to be used for broker authentication',
        'type': 'string',
        'default': '',
        'order': '15',
        'displayName': 'Username',
        'validity': "authentication == \"SASL_PLAINTEXT\" || authentication == \"SASL_SSL\"",
        'group': 'Kafka Authentication'
    },
    'password': {
        'description': 'Password to be used for broker authentication',
        'type': 'password',
        'default': '',
        'order': '16',
        'displayName': 'Password',
        'validity': "authentication == \"SASL_PLAINTEXT\" || authentication == \"SASL_SSL\"",
        'group': 'Kafka Authentication'
    },
    'ssl_cafile': {
        'description': 'Name of Root CA to use in certificate verification.',
        'type': 'string',
        'default': '',
        'order': '17',
        'displayName': 'Root CA Name',
        'validity': "authentication == \"SSL\" || authentication == \"SASL_SSL\"",
        'group': 'Kafka SSL'
    },
    'ssl_certfile': {
        'description': 'Name of client certificate for identity authentications.',
        'type': 'string',
        'default': '',
        'order': '18',
        'displayName': 'Certificate Name',
        'validity': "authentication == \"SSL\" || authentication == \"SASL_SSL\"",
        'group': 'Kafka SSL'
    },
    'ssl_keyfile': {
        'description': 'Name of client private key required for communication.',
        'type': 'string',
        'default': '',
        'order': '19',
        'displayName': 'Private Key Name',
        'validity': "authentication == \"SSL\" || authentication == \"SASL_SSL\"",
        'group': 'Kafka SSL'
    },
    'ssl_password': {
        'description': 'Optional: Password to be used when loading the certificate chain.',
        'type': 'password',
        'default': '',
        'order': '20',
        'displayName': 'SSL Certificate Password',
        'validity': "authentication == \"SSL\" || authentication == \"SASL_SSL\"",
        'group': 'Kafka SSL'
    }
}


def plugin_info():
    return {
        'name': 'Kafka Consumer Plugin',
        'version': '3.0.0',
        'mode': 'async',
        'type': 'south',
        'interface': '2.0',
        'config': _DEFAULT_CONFIG
    }


def plugin_init(config):
    """Registers Kafka Consumer Client

    Args:
        config: JSON configuration document for the South plugin configuration category
    Returns:
        handle: JSON object to be used in future calls to the plugin
    Raises:
    """
    handle = copy.deepcopy(config)
    handle["_kafka"] = KafkaConsumerClient(handle)
    return handle


def plugin_start(handle):
    """ Starts the plugin
    """
    
    _LOGGER.info('Starting kafka_python south plugin...')
    try:
        _kafka = handle["_kafka"]
        _kafka.start()
    except Exception as e:
        _LOGGER.exception(str(e))
    else:
        _LOGGER.info('Kafka python south plugin started.')
    
def plugin_shutdown(handle):
    """ Shut down the plugin

    To be called prior to the South service being shut down.

    Args:
        handle: handle returned by the plugin initialisation call
    Returns:
    Raises:
    """
    
    try:
        _LOGGER.debug('Shutting down kafka python south plugin...')
        _kafka = handle["_kafka"]
        _kafka.stop()
        _kafka.event_loop.run_until_complete(asyncio.sleep(1))
        _kafka.event_loop.close()
        _kafka.event_loop = None
        
    except Exception as e:
        _LOGGER.error('Failed to shut down: {}'.format(str(e)))
    else:
        _LOGGER.info('Kafka south plugin shut down.')


def plugin_register_ingest(handle, callback, ingest_ref):
    """ Required plugin interface component to communicate to South C server
    Args:
        handle: handle returned by the plugin initialisation call
        callback: C opaque object required to passed back to C->ingest method
        ingest_ref: C opaque object required to passed back to C->ingest method
    """
    global c_callback, c_ingest_ref
    c_callback = callback
    c_ingest_ref = ingest_ref

def plugin_reconfigure(handle, new_config):
    """ Reconfigures the plugin

    it should be called when the configuration of the plugin is changed during the operation of the South service;
    The new configuration category should be passed.

    Args:
        handle: handle returned by the plugin initialisation call
        new_config: JSON object representing the new configuration category for the category
    Returns:
        new_handle: new handle to be used in the future calls
    Raises:
    """
    _LOGGER.info('Reconfiguring Kafka south plugin...')
    plugin_shutdown(handle)

    new_handle = plugin_init(new_config)
    plugin_start(new_handle)

    _LOGGER.info('Kafka python south plugin reconfigured.')
    return new_handle

class KafkaConsumerClient(object):
    """ kafka listener class"""

    __slots__ = ['broker', 'topic_assets_map', 'authentication', 'asset', 'topic_list', 'stop_connection',
                 'kafka_config', 'kafka_consumer', 'message_thread', 'event_loop', 'poll_timeout']
    
    def __init__(self, config):
        self.broker = "{}:{}".format(config['brokerHost']['value'], int(config['brokerPort']['value']))
        self.topic_assets_map = config['topic_assets_map']['value']['map']
        self.authentication = config['authentication']['value']
        self.poll_timeout = int(config['timeout_ms']['value'])  # Message Poll timeout in milliseconds
        
        if self.authentication in ['GSSAPI', 'OAUTHBEARER', "SCRAM-SHA-256", "SCRAM-SHA-512"]:
            _LOGGER.warning("{} is not supported yet".format(self.authentication))
        
        self.asset = config['asset']['value']
        if self.topic_assets_map != []:
            self.topic_list = [ ele['topic'] for ele in self.topic_assets_map if 'topic' in ele and ele['topic'].strip() != '']
            _LOGGER.info("Topic List: {}".format(self.topic_list ))
            if len(self.topic_list) == 0:
                _LOGGER.error("Topic is missing")
        else:
            _LOGGER.error("Topic asset JSON map missing")
        
        # Kafka Consumer Configuration
        self.set_kafka_config(config)
        self.kafka_consumer = None
        
        self.message_thread = None
        # Variable to stop kafka consumer connection        
        self.stop_connection = False
        
        try:
            # Get current event loop
            self.event_loop = asyncio.get_event_loop()
        except RuntimeError as e:
            # Create event loop
            self.event_loop = asyncio.new_event_loop()
            asyncio.set_event_loop(self.event_loop)
    
    def set_kafka_config(self, config):
        """
            Helper funtion to set configuration of kafka consumer
        """ 
               
        # Get Certificate Storage Path
        cert_storage_path = self.get_certs_dir()
        
        self.kafka_config = dict()
        self.kafka_config['bootstrap_servers'] = self.broker
        self.kafka_config['group_id'] = config['group_id']['value']
        self.kafka_config['auto_offset_reset'] = config['auto_offset_reset']['value']
        self.kafka_config['enable_auto_commit'] = True if config['enable_auto_commit']['value'] == 'true' else False
        self.kafka_config['heartbeat_interval_ms'] = int(config['heartbeat_interval_ms']['value'])
        self.kafka_config['request_timeout_ms'] = int(config['request_timeout_ms']['value'])
        self.kafka_config['reconnect_backoff_ms'] = int(config['reconnect_backoff_ms']['value'])
        self.kafka_config['reconnect_backoff_max_ms'] = int(config['reconnect_backoff_max_ms']['value'])
        
        # For Authentication
        if self.authentication != "PLAINTEXT":
            self.kafka_config['security_protocol'] = self.authentication
            
        if self.authentication == "SASL_PLAINTEXT" or self.authentication == "SASL_SSL":
            self.kafka_config['sasl_mechanism'] = config['sasl_mechanism']['value']
            self.kafka_config['sasl_plain_username'] = config['username']['value']
            self.kafka_config['sasl_plain_password'] = config['password']['value']

        # For Consuming messages from Azure Event Hub using SAS mechanism. For more detials: 
        # https://learn.microsoft.com/en-us/azure/event-hubs/azure-event-hubs-kafka-overview#shared-access-signature-sas
        if (config['username']['value'] != "$ConnectionString") and (self.authentication == "SSL" or self.authentication == "SASL_SSL"):
            self.kafka_config['ssl_cafile'] = cert_storage_path + config['ssl_cafile']['value']
            self.kafka_config['ssl_certfile'] = cert_storage_path + config['ssl_certfile']['value']
            self.kafka_config['ssl_keyfile'] = cert_storage_path + config['ssl_keyfile']['value']
            if config['ssl_password']['value'] != '':
                self.kafka_config['ssl_password'] = config['ssl_password']['value']
    
    def get_certs_dir(self):
        """ Helper function to fetch FOGLAMP_DATA directory path.
            Returns:
                Directory path of stored certificates.
        """
        dir_path = _FOGLAMP_DATA + '/etc/certs/' if _FOGLAMP_DATA else _FOGLAMP_ROOT + '/data/etc/certs/'
        if not os.path.exists(dir_path):
            os.makedirs(dir_path)
        certs_dir = os.path.expanduser(dir_path)
        return certs_dir
    
    async def consume_messages(self):
        try:
            _LOGGER.debug("Fetching data from Kafka Broker")
            while not self.stop_connection:
                messages = self.kafka_consumer.poll(timeout_ms=self.poll_timeout)
                for topic_partition, message in messages.items():
                    for record in message:
                        await self.save(record.topic, record.value)
        
        except Exception as ex:
            _LOGGER.exception('Error consuming messages from Kafka: {}'.format(ex))
            raise ex
    
    def start(self):
        """ To start Kafka Client
        """
        try:
            _LOGGER.debug("Kafka Configuration: {}".format(self.kafka_config))
            self.kafka_consumer = KafkaConsumer(**self.kafka_config)
            if self.kafka_consumer.bootstrap_connected():
                # Subscribe to the topic
                self.kafka_consumer.subscribe(self.topic_list)
                _LOGGER.info("Connected to Kafka broker and Subscribed to topics: {}".format(self.topic_list))
                self.message_thread = threading.Thread(target=self.event_loop.run_until_complete, args=(self.consume_messages(),))
                self.message_thread.start()
            else:
                _LOGGER.error("Failed to connect to Kafka Broker: {}".format(self.broker))
        except Exception as ex:
            _LOGGER.error('Failed to start Kafka consumer: {}'.format(ex))
        

    def stop(self):
        _LOGGER.debug("stop called")
        if self.kafka_consumer is not None:
            self.stop_connection = True
            if self.message_thread is not None:
                self.message_thread.join()
            self.kafka_consumer.close()

    async def save(self, source, msg):
        """Store msg content to FogLAMP """
        global c_callback, c_ingest_ref
        payload_json = json.loads(msg.decode('utf-8'))
        # _LOGGER.info("PAYLOAD_JSON: {}".format(payload_json))
        
        asset_name = self.asset
        dp_name = ""
        found = False
        
        for item in self.topic_assets_map:
            if source == item['topic']:
                found = True
                if item['asset'].strip() != "":
                    asset_name = item['asset']
                    
                if item['datapoint'].strip() == "":
                    dp_name = item['topic']
                else:
                    dp_name = item['datapoint']
                   
        if not found:
            _LOGGER.error("Unknown Topic: {}".format(source))
            return
        
        data = {
            'timestamp': utils.local_timestamp(),
            'asset': asset_name,
            'readings': {dp_name: '{}'.format(payload_json)}
        }                    
        _LOGGER.debug("payload dict --> {}".format(data))
        async_ingest.ingest_callback(c_callback, c_ingest_ref, data)
